from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.base import MultiOutputMixin
from sklearn.utils.validation import check_is_fitted
from abc import abstractmethod
import gurobipy as gp
from .objective import Objective, ObjectivePart, BaseObjective
from copy import copy
import json


class ModelBuilder:
    """This should be user implemented"""

    def __init__(self, data):
        self.data = data
        self.objective = None

    @abstractmethod
    def build_variables(self, base_model: gp.Model) -> None:
        raise NotImplementedError(
            f"{type(self).__name__} should implement build_variables!"
        )

    @abstractmethod
    def build_constraints(self, base_model: gp.Model) -> None:
        raise NotImplementedError(
            f"{type(self).__name__} should implement build_constraints!"
        )

    @abstractmethod
    def build_objective(self, base_model) -> Objective:
        raise NotImplementedError(
            f"{type(self).__name__} should implement build_objective!"
        )

    def build(self):
        base_model = gp.Model("my_model")
        self.build_variables(base_model)
        base_model.update()
        self.build_constraints(base_model)
        base_model.update()
        self.objective = self.build_objective(base_model)
        base_model.update()
        return base_model


def model_builder_factory(variables_builder, constraints_builder, objective_builder):
    class ModelBuilderF(ModelBuilder):
        def build_variables(self, base_model: gp.Model) -> None:
            variables_builder(base_model, self.data)

        def build_constraints(self, base_model: gp.Model) -> None:
            constraints_builder(base_model, self.data)

        def build_objective(self, base_model: gp.Model) -> Objective:
            return objective_builder(base_model, self.data)

    return ModelBuilderF


class OptModel:  # (MultiOutputMixin, RegressorMixin, BaseEstimator):
    vars_: dict
    objective_value_: float
    fit_callback_data: dict

    def __init__(self, *, model_builder):
        self.ModelBuilder = model_builder
        self.data = None
        self.objective = None

    def fit(self, data, callback=None):
        self.data = data
        """Builds and optimize the specific the model given the data"""
        # TODO: add some checks over data here may be feasibility
        model_builder = self.ModelBuilder(data)
        model = model_builder.build()
        model.optimize()

        self.objective = json.loads(model_builder.objective.__repr__())
        try:
            self.vars_ = {v.var_name: v.x for v in model.getVars()}
            self.objective_value_ = model.getObjective().getValue()
            if callback:
                self.fit_callback_data = callback(model)
        except Exception:
            del self.vars_, self.objective_value_, self.fit_callback_data
        finally:
            del model
        return self

    def predict(self, data, callback=None):
        """Fits estimator if not fitted or self.data differs from data and returns the variable values"""
        fitted = [v for v in vars(self) if v.endswith("_") and not v.startswith("__")]
        if not fitted or self.data != data:
            self.fit(data, callback=callback)
        return self.vars_

    def score(self):
        """Returns the spefic model objective given the data"""
        check_is_fitted(self)
        return self.objective_value_



from typing import Union, List, Iterable
import json


def pretty_json_str(json_str):
    return json.dumps(json.loads(json_str.replace("'", '"')), indent=2, sort_keys=True)


def is_jsonable(x):
    try:
        json.dumps(x)
        return True
    except (TypeError, OverflowError):
        return False


class ObjectivePart:
    def __init__(self, weight: float, expr):
        self._weight = weight
        self._expr = expr

    def __add__(self, other):
        expr = self._weight * self._expr + other._weight * other._expr
        return ObjectivePart(weight=1, expr=expr)

    def __repr__(self):
        return json.dumps(
            {k: v if is_jsonable(v) else str(v) for k, v in self.__dict__.items()},
            indent=2,
            sort_keys=True,
        )


class BaseObjective:
    def __init__(self, objective_parts: List[ObjectivePart], hierarchy: int):
        self._objective_parts = objective_parts
        self._hierarchy = hierarchy

    def build(self):
        objective_parts_sum = sum(
            (objective_part for objective_part in self._objective_parts),
            ObjectivePart(weight=1, expr=0),
        )
        return objective_parts_sum._expr

    def __repr__(self):
        return pretty_json_str(str(self.__dict__))


class Objective:
    def __init__(self, base_objectives: Iterable[BaseObjective]):
        self._base_objectives = base_objectives

    def build(self) -> Union[BaseObjective, List[BaseObjective]]:
        try:
            return [base_objective.build() for base_objective in self._base_objectives]
        except TypeError:
            return self._base_objectives.build()

    def __repr__(self):
        return pretty_json_str(str(self.__dict__))



from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.rst").read_text()

setup(
    name="opt_sugar",
    version="0.0.4.3",
    author="Juan Chacon",
    author_email="juandados@gmail.com",
    packages=find_packages("src"),
    package_dir={"": "src"},
    url="https://github.com/juandados/opt-sugar",
    keywords="optimization operations mathematical programming",
    install_requires=[
        "numpy",
        "gurobipy",
        "scikit-learn",
    ],
    long_description=long_description,
    long_description_content_type='text/x-rst'
)
